
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <signal.h>
#include <time.h>
#include <string.h>

#define MAXTREES 50000
#define INPUTFILENAME "c:\\harp_input.txt"
#define CUTOUTPUTFILENAME "c:\\harp_cut_output.txt"
#define SAVEDOUTPUTFILENAME "c:\\harp_saved_output.txt"
#define MINCUTDIAM 12.0
#define SKIDTRAILWIDTH 6.0
#define HARVESTAREAWIDTH 8.0
#define STUMPCOUNTRADIUS 7.35
#define PLOTXLENGTH 224
#define PLOTYLENGTH 224
#define PI 3.1416
#define HARVESTYEAR 1

//Note: min diameter of adults should be set to 5 cm (smallest trees that can be stumps)

typedef struct {
  float X;
  float Y;
  int Species;
  int Type;
  float Diam;
  float Height;
  int NearbyStumpCount;
  int oldstumpcount;
  bool OnSkidTrail;
  int StripNum;
  int SortOrder;
} Tree;

Tree InputTreeList[MAXTREES], //trees in the input file
      CutTreeList[MAXTREES], //trees getting cut
      ResidualTreeList[MAXTREES]; //trees not getting cut

int nInputTrees,
    nCutTrees,
    nResidualTrees,
    MaxStripNum,
    ResidualStripIndices[100];

char FirstLine[100];
char SecondLine[100];

int timestep;
int timestep2;

void ReadFile();
void ApplyHarvest();
void CountNearbyStumps();
void WriteFiles();

int CompStripNum(const void *a, const void *b);
int CompSortOrder(const void *a, const void *b);

int main() {

  //read inputs
  ReadFile();

  //sort by strip number
  qsort(InputTreeList, nInputTrees, sizeof(Tree), CompStripNum);

  //copy trees to cut and residual tree lists as per HARP
  ApplyHarvest();

  //tally the number of skid-trail stumps near each residual tree
  CountNearbyStumps();

  //sort back to original order
  qsort(CutTreeList, nCutTrees, sizeof(Tree), CompSortOrder);
  qsort(ResidualTreeList, nResidualTrees, sizeof(Tree), CompSortOrder);

  //write output files for cut and residual tree lists
  WriteFiles();

  return 0;
}

/**************************************************************************/

void ReadFile() {

  FILE *ifp;

  float X,Y,Diam,Height,
    StripWidth = HARVESTAREAWIDTH + SKIDTRAILWIDTH;
  int Species, Type, oldstumpcount;
  char line[100];

  ifp = fopen(INPUTFILENAME, "r");
    if (ifp==NULL) {
      fprintf(stderr,"Can't open input file\n");
      exit(99);
    }

//scan input data

  //first line is timestep info
  fgets(FirstLine, 100, ifp);
  sscanf(FirstLine, "%d\t%d", &timestep, &timestep2);

  //second line is the column headings
  fgets(SecondLine, 100, ifp);

  //subsequent lines are the trees
  nInputTrees=0;

  while (fgets(line, 100, ifp)) {

    sscanf(line,"%f\t%f\t%d\t%d\t%f\t%f\%d\n",
        &X,
        &Y,
        &Species,
        &Type,
        &Diam,
        &Height,
        &oldstumpcount);

    //assign data to structure
    InputTreeList[nInputTrees].X = X;
    InputTreeList[nInputTrees].Y = Y;
    InputTreeList[nInputTrees].Species = Species;
    InputTreeList[nInputTrees].Type = Type;
    InputTreeList[nInputTrees].Diam = Diam;
    InputTreeList[nInputTrees].Height = Height;
    InputTreeList[nInputTrees].oldstumpcount = oldstumpcount;

    InputTreeList[nInputTrees].StripNum =
        (int)(X/StripWidth);

    InputTreeList[nInputTrees].OnSkidTrail =
        ((X - InputTreeList[nInputTrees].StripNum*StripWidth) > HARVESTAREAWIDTH);

    InputTreeList[nInputTrees].SortOrder=nInputTrees;

    nInputTrees++;

    if (nInputTrees == MAXTREES) {

      fprintf(stderr, "Number of input trees exceeds max length of array!\n");
      exit(99);

    } //end if

  } //end while

  fclose(ifp);

  return;

}

/**************************************************************************/

void ApplyHarvest() {

  int i,
      LastResidualStripNum=0;

  nCutTrees=0;
  nResidualTrees=0;

  ResidualStripIndices[0]=0;

///////////////////////IF A HARVEST YEAR/////////////////////////////////

if (timestep == HARVESTYEAR) {

  for (i=0; i<nInputTrees; i++) {

    //partial harvest areas are HARVESTAREAWIDTH m wide, starting on the left
    //edge of the plot; skid trails are SKIDTRAILWIDTH m wide and alternate with
    //partial harvest strips


    if (InputTreeList[i].OnSkidTrail //either it's on the skid trail
        || ((InputTreeList[i].Type == 3) && (InputTreeList[i].Diam > MINCUTDIAM))) {
        //or it exceeds the diameter limit

    //the tree is going to be cut

      //copy tree to the cut tree list
      CutTreeList[nCutTrees].X = InputTreeList[i].X;
      CutTreeList[nCutTrees].Y = InputTreeList[i].Y;
      CutTreeList[nCutTrees].Species = InputTreeList[i].Species;
      CutTreeList[nCutTrees].Type = InputTreeList[i].Type;
      CutTreeList[nCutTrees].Diam = InputTreeList[i].Diam;
      CutTreeList[nCutTrees].Height = InputTreeList[i].Height;

      CutTreeList[nCutTrees].OnSkidTrail = InputTreeList[i].OnSkidTrail;
      CutTreeList[nCutTrees].StripNum = InputTreeList[i].StripNum;
      CutTreeList[nCutTrees].SortOrder = InputTreeList[i].SortOrder;

      //increment the number of cut trees
      nCutTrees++;

    } //end if cut

    else { //not cut

      //copy tree to the residual tree list
      ResidualTreeList[nResidualTrees].X = InputTreeList[i].X;
      ResidualTreeList[nResidualTrees].Y = InputTreeList[i].Y;
      ResidualTreeList[nResidualTrees].Species = InputTreeList[i].Species;
      ResidualTreeList[nResidualTrees].Type = InputTreeList[i].Type;
      ResidualTreeList[nResidualTrees].Diam = InputTreeList[i].Diam;
      ResidualTreeList[nResidualTrees].Height = InputTreeList[i].Height;

      ResidualTreeList[nResidualTrees].NearbyStumpCount = 0;

      ResidualTreeList[nResidualTrees].StripNum = InputTreeList[i].StripNum;
      ResidualTreeList[nResidualTrees].SortOrder = InputTreeList[i].SortOrder;

      //if this is a new strip number for residual trees, write the index value
      if (ResidualTreeList[nResidualTrees].StripNum > LastResidualStripNum) {
        ResidualStripIndices[ResidualTreeList[nResidualTrees].StripNum] = nResidualTrees;
        LastResidualStripNum = ResidualTreeList[nResidualTrees].StripNum;
      }

      //increment the number of residual trees
      nResidualTrees++;

    } //end else not cut


  //record keeping for the last strip indices
  MaxStripNum = InputTreeList[nInputTrees].StripNum;
  ResidualStripIndices[MaxStripNum+1] = nResidualTrees;
  ResidualStripIndices[MaxStripNum+2] = nResidualTrees;

  } //end for (input trees)
  } //end IF HARVESTYEAR

//////////IF NOT A HARVEST YEAR///////////////////////////

if(timestep != HARVESTYEAR) {

  for (i=0; i<nInputTrees; i++) {

      //copy tree to the residual tree list
      ResidualTreeList[nResidualTrees].X = InputTreeList[i].X;
      ResidualTreeList[nResidualTrees].Y = InputTreeList[i].Y;
      ResidualTreeList[nResidualTrees].Species = InputTreeList[i].Species;
      ResidualTreeList[nResidualTrees].Type = InputTreeList[i].Type;
      ResidualTreeList[nResidualTrees].Diam = InputTreeList[i].Diam;
      ResidualTreeList[nResidualTrees].Height = InputTreeList[i].Height;

      ResidualTreeList[nResidualTrees].NearbyStumpCount = InputTreeList[i].oldstumpcount;

      ResidualTreeList[nResidualTrees].StripNum = InputTreeList[i].StripNum;
      ResidualTreeList[nResidualTrees].SortOrder = InputTreeList[i].SortOrder;

      //if this is a new strip number for residual trees, write the index value
      if (ResidualTreeList[nResidualTrees].StripNum > LastResidualStripNum) {
        ResidualStripIndices[ResidualTreeList[nResidualTrees].StripNum] = nResidualTrees;
        LastResidualStripNum = ResidualTreeList[nResidualTrees].StripNum;
      }

      //increment the number of residual trees
      nResidualTrees++;


  //record keeping for the last strip indices
  MaxStripNum = InputTreeList[nInputTrees].StripNum;
  ResidualStripIndices[MaxStripNum+1] = nResidualTrees;
  ResidualStripIndices[MaxStripNum+2] = nResidualTrees;

  } //end for (input trees)
  } //end if timestep !=HARVESTYEAR

  return;

}

/**************************************************************************/

void CountNearbyStumps() {

  int i, j;

  float Xdist, Ydist;


///////IF HARVEST YEAR//////////
  if (timestep == HARVESTYEAR){

  for (i=0; i<nCutTrees; i++) {

    if (CutTreeList[i].OnSkidTrail && (CutTreeList[i].Type == 3)) {
      //this is a skid trail stump

      for (j=ResidualStripIndices[CutTreeList[i].StripNum];
            j<ResidualStripIndices[CutTreeList[i].StripNum+2];
            j++) {

        //loop through the residual trees in adjacent partial cut strips

        //find distances in x and y directions
        Xdist = fabs(CutTreeList[i].X - ResidualTreeList[j].X);

        Ydist = fabs(CutTreeList[i].Y - ResidualTreeList[j].Y);
        if (Ydist > PLOTYLENGTH/2)
          Ydist = PLOTYLENGTH - Ydist;

        //if the diagonal distance is less than the stump count radius,
        //then increment the stump count for this tree
        if (pow(pow(Xdist,2) + pow(Ydist,2),0.5) < STUMPCOUNTRADIUS)
            ResidualTreeList[j].NearbyStumpCount++;

      }//end for each residual tree in adjacent strips

      if (CutTreeList[i].StripNum > (MaxStripNum - 2)) {
      //if the stump is near the right edge of the plot, check the trees near
      //the left edge as well for wrapping around the plot torus
      for (j=ResidualStripIndices[0];
            j<ResidualStripIndices[1];
            j++) {

        //find distances in x and y directions
        Xdist = PLOTXLENGTH - fabs(CutTreeList[i].X - ResidualTreeList[j].X);

        Ydist = fabs(CutTreeList[i].Y - ResidualTreeList[j].Y);
        if (Ydist > PLOTYLENGTH/2)
          Ydist = PLOTYLENGTH - Ydist;


        //if the diagonal distance is less than the stump count radius,
        //then increment the stump count for this tree
        if (pow(pow(Xdist,2) + pow(Ydist,2),0.5) < STUMPCOUNTRADIUS)
            ResidualTreeList[j].NearbyStumpCount++;

      }//end for (each residual tree in the first strip)
      }//end if (the stump is near the right edge of the plot)

    } //end if (this is a skid trail stump)


  } //end for (cut tree list)

} //end if HARVESTYEAR


  return;

}

/**************************************************************************/

void WriteFiles() {

  int i;
  float BA;
  FILE *ofp1, *ofp2, *ofp3, *ofp4;

////////////IF HARVEST YEAR///////////////

if (timestep == HARVESTYEAR){

  ofp1 = fopen(CUTOUTPUTFILENAME,"w");
  if (ofp1==NULL) {
    printf("Can't open cut output file.\n");
    exit(99);
  }

  ofp3 = fopen("c:/cuttrees.txt", "w");
  if (ofp3==NULL) {
    printf("Can't open cut output file.\n");
    exit(99);
  }

  fputs(FirstLine, ofp1);
  fputs(SecondLine, ofp1);

  for(i=0;i<nCutTrees;i++) {

        BA = PI * (CutTreeList[i].Diam/200.00) * (CutTreeList[i].Diam/200.0);

        fprintf(ofp3,"%4.4f\t%4.4f\t%4.4f\t%4.4f\n",
          CutTreeList[i].X,
          CutTreeList[i].Y,
          CutTreeList[i].Diam,
          BA);

        fprintf(ofp1,"%4.4f\t%4.4f\t%d\t%d\t%4.4f\t%4.4f\t%d\n",
          CutTreeList[i].X,
          CutTreeList[i].Y,
          CutTreeList[i].Species,
          CutTreeList[i].Type,
          CutTreeList[i].Diam,
          CutTreeList[i].Height,
          CutTreeList[i].SortOrder);

  } //end for

  fclose(ofp1);
  fclose(ofp3);
} //end if

////////////////IF NO HARVEST//////////////////////////////////

 if (timestep != HARVESTYEAR){

  ofp1 = fopen(CUTOUTPUTFILENAME,"w");
  if (ofp1==NULL) {
    printf("Can't open cut output file.\n");
    exit(99);
  }

  fputs(FirstLine, ofp1);
  fputs(SecondLine, ofp1);

  for(i=0;i<nCutTrees;i++) {

        fprintf(ofp1,"%4.4f\t%4.4f\t%d\t%d\t%4.4f\t%4.4f\t%d\n",
          CutTreeList[i].X,
          CutTreeList[i].Y,
          CutTreeList[i].Species,
          CutTreeList[i].Type,
          CutTreeList[i].Diam,
          CutTreeList[i].Height,
          CutTreeList[i].SortOrder);

  } //end for

  fclose(ofp1);
} //end if

/////////////////////////////////////////////////////////////////////


/////////IF HARVEST YEAR////////////

if (timestep == HARVESTYEAR){

  ofp2 = fopen(SAVEDOUTPUTFILENAME,"w");
  if (ofp2==NULL) {
    printf("Can't open saved output file.\n");
    exit(99);
  }

  ofp4 = fopen("c:/residtrees.txt", "w");
  if (ofp4==NULL) {
    printf("Can't open cut output file.\n");
    exit(99);
  }

  fputs(FirstLine, ofp2);
  fputs(SecondLine, ofp2);

  for(i=0;i<nResidualTrees;i++) {

       BA = PI * (ResidualTreeList[i].Diam/200.00) * (ResidualTreeList[i].Diam/200.0);

      fprintf(ofp2,"%4.4f\t%4.4f\t%d\t%d\t%4.4f\t%4.4f\t%d\n",
        ResidualTreeList[i].X,
        ResidualTreeList[i].Y,
        ResidualTreeList[i].Species,
        ResidualTreeList[i].Type,
        ResidualTreeList[i].Diam,
        ResidualTreeList[i].Height,
        ResidualTreeList[i].NearbyStumpCount);

        fprintf(ofp4,"%4.4f\t%4.4f\t%4.4f\t%4.4f\n",
          ResidualTreeList[i].X,
          ResidualTreeList[i].Y,
          ResidualTreeList[i].Diam,
          BA);


  } //end for

  fclose(ofp2);
  fclose(ofp4);

  } //end if HARVESTYEAR



 //////////IF NOT HARVESTYEAR//////////////////

if (timestep != HARVESTYEAR){

  ofp2 = fopen(SAVEDOUTPUTFILENAME,"w");
  if (ofp2==NULL) {
    printf("Can't open saved output file.\n");
    exit(99);
  }

  fputs(FirstLine, ofp2);
  fputs(SecondLine, ofp2);

  for(i=0;i<nResidualTrees;i++) {

      fprintf(ofp2,"%4.4f\t%4.4f\t%d\t%d\t%4.4f\t%4.4f\t%d\n",
        ResidualTreeList[i].X,
        ResidualTreeList[i].Y,
        ResidualTreeList[i].Species,
        ResidualTreeList[i].Type,
        ResidualTreeList[i].Diam,
        ResidualTreeList[i].Height,
        ResidualTreeList[i].NearbyStumpCount);

    } //end for

  fclose(ofp2);
} //end if not HARVESTYEAR

////////////////////////////////////////////////////

  return;

}

/**************************************************************************/

int CompStripNum(const void *a, const void *b) {

  if ((*(Tree*)a).StripNum < (*(Tree*)b).StripNum)
    return -1;
  else if ((*(Tree*)a).StripNum > (*(Tree*)b).StripNum)
    return 1;
  else
    return 0;

}

/**************************************************************************/

int CompSortOrder(const void *a, const void *b) {

  return ((*(Tree*)a).SortOrder - (*(Tree*)b).SortOrder);

}

/**************************************************************************/

